
import numpy as np
import torch
import torch.nn.functional as F
from scipy.sparse import csr_matrix

# Most of code in this script is borrowed from https://github.com/zhoujt1994/scHiCluster.

def convolution_p(rawHiC, pad=1, device=torch.device('cuda:0')):
	if pad == 0:
		return torch.from_numpy(rawHiC).float().to(device)
	kernel_size = 2*pad + 1
	conv_filter = torch.ones(1, 1, kernel_size, kernel_size).to(device)
	convHiC = F.conv2d(torch.from_numpy(rawHiC[None,None,:,:]).float().to(device),
                           conv_filter, padding=2*pad)
	return (convHiC[0,0,pad:-pad,pad:-pad] / float(kernel_size*kernel_size))

def convolution_p2(rawHiC, pad=1, device=torch.device('cuda:0')):
	if pad == 0:
		return torch.from_numpy(rawHiC).float().to(device)
	kernel_size = 2*pad + 1
	conv_filter = torch.ones(1, 1, kernel_size, kernel_size).to(device)
	convHiC = F.conv2d(torch.from_numpy(rawHiC[None,None,:,:]).float().to(device),
                           conv_filter, padding=2*pad)
	return convHiC[0,0,pad:-pad,pad:-pad].cpu().data.numpy()


def random_walk(convHiC, rp=0.5, device=torch.device('cuda:0')):
	# revised scHiCluster, which makes the output matrix symmetric
	nbins = convHiC.shape[0]
	convHiC.fill_diagonal_(0.0)
	#convHiC = convHiC + torch.diag(torch.sum(convHiC,0)==0).float()
	convHiC[range(nbins), range(nbins)] = (torch.sum(convHiC,0) == 0).type(torch.float)
	
	#P = torch.div(convHiC, torch.sum(convHiC, 0))
	# scHiCluster-symm
	rsum = torch.diag(torch.pow(torch.sum(convHiC,0), -0.5))
	P = torch.mm(torch.mm(rsum,convHiC),rsum)

	Q = torch.eye(nbins).to(device)
	I = torch.eye(nbins).to(device)
	for i in range(30):
		Q_new = (1-rp)*I + rp*torch.mm(Q, P)
		delta = torch.norm(Q-Q_new, 2)
		Q = Q_new
		if delta < 1e-6:
			break
	return Q


def scHiCluster(a, top=20, pad=1, 
                device=torch.device('cuda:0' if torch.cuda.is_available() else "cpu")):
	'''
	input a: raw single-cell symmetric matrix (numpy)
	output f: binary matrix (numpy) - top as one, zero otherwise
	'''
	n = a.shape[0]
	# step 1: convolution imputation
	b = np.log2(a + 1)	
	b = convolution_p(b, pad = pad, device = device)
	# step 2: RWR imputation
	c = random_walk(b, device = device)
	d = c.cpu().numpy()
	# step 3: top 
	e = np.percentile(d, 100 - top)
	f = np.zeros((n, n), dtype = int)
	f[d > e] = 1
	return f


